#include <kernel/kernel.h>
#include <kernel/multiboot.h>
#include <kernel/acpi.h>
#include <kernel/vga.h>
#include <kernel/pci.h>
#include <kernel/printf.h>
#include <kernel/strcmp.h>
#include <kernel/gdt.h>
#include <kernel/idt.h>
#include <kernel/pic8259.h>
#include <kernel/ps2.h>
#include <kernel/ide.h>
#include <kernel/mem.h>

#define MAGIC_BREAKPOINT __asm__ volatile("xchgw %bx, %bx");
#define FD_STDOUT 0
#define FD_STDIN  1

struct kernel_context ctx = {0};
struct idt_entry g_idt[256] = {0};
struct idtr g_idtr = {0};
extern struct gdt_entry gdt[6];
extern void syscall();
extern void test();

struct tss_entry tss;

struct stack_frame {
    uint32_t edi, esi, ebp, esp, ebx, edx, ecx, eax;
};

void syscall_read(struct stack_frame regs) {
    uint32_t fd    = regs.edi;
    char *buf      = (char *)regs.esi;
    uint32_t count = regs.edx;

    if (fd == FD_STDIN) {
        extern char shell_buffer[512];
        extern int shell_index;
        asm volatile ("sti" ::);

        while (shell_buffer[shell_index-1] != '\n')
            asm volatile ("pause" ::);
        asm volatile ("cli" ::);

        memcpy(buf, shell_buffer, count);
        memset(shell_buffer, 0, 512);
        shell_index = 0;
    }
}

void syscall_write(struct stack_frame regs) {
    uint32_t fd    = regs.edi;
    char *buf      = (char *)regs.esi;
    //uint32_t count = regs.edx;

    if (fd == FD_STDOUT) {
        printf("%s", buf);
    }
}

void switch_to_user_mode(void *user_stack, void *user_entry) {
    asm volatile (
        "movl $0x23, %%eax \n"      // User data segment selector
        "mov %%ax, %%ds \n"
        "mov %%ax, %%es \n"
        "mov %%ax, %%fs \n"
        "mov %%ax, %%gs \n"

        "pushl $0x23 \n"            // User stack segment
        "pushl %0 \n"               // User stack pointer
        "pushfl \n"                 // Push EFLAGS
        "popl %%eax \n"
        "orl $0x200, %%eax \n"      // Enable interrupts (set IF)
        "pushl %%eax \n"

        "pushl $0x1B \n"            // User code segment
        "pushl %1 \n"               // User entry point

        "iret \n"                   // Interrupt return (switches to user mode)
        :
        : "r"(user_stack), "r"(user_entry)
        : "eax"
    );
}

void user_init() {
// Setup syscalls
    uint32_t eip = (uint32_t)&syscall;
    uint32_t esp;

    // Get the current value of ESP
    __asm__ volatile ("mov %%esp, %0" : "=r"(esp));

    // Write to MSR_SYSENTER_CS (0x174)
    __asm__ volatile (
        "movl $0x174, %%ecx \n\t"   // MSR address
        "movl %0, %%eax \n\t"       // Lower 32 bits (cs_selector)
        "xor %%edx, %%edx \n\t"     // Upper 32 bits (zero for 32-bit value)
        "wrmsr"
        : 
        : "r"(GDT_SEGMENT_SELECTOR(GDT_KERNEL_CODE, RING0))
        : "eax", "ecx", "edx"
    );

    // Write to MSR_SYSENTER_ESP (0x175)
    __asm__ volatile (
        "movl $0x175, %%ecx \n\t"   // MSR address
        "movl %0, %%eax \n\t"       // Lower 32 bits (esp)
        "xor %%edx, %%edx \n\t"     // Upper 32 bits (zero for 32-bit value)
        "wrmsr"
        : 
        : "r"(esp)
        : "eax", "ecx", "edx"
    );

    // Write to MSR_SYSENTER_EIP (0x176)
    __asm__ volatile (
        "movl $0x176, %%ecx \n\t"   // MSR address
        "movl %0, %%eax \n\t"       // Lower 32 bits (eip)
        "xor %%edx, %%edx \n\t"     // Upper 32 bits (zero for 32-bit value)
        "wrmsr"
        : 
        : "r"(eip)
        : "eax", "ecx", "edx"
    );

// Switch
    write_tss(&tss, gdt, 5, 0x10, 0x0);
    tss_flush(0x2B);
    void *page = (uint32_t *)0x6000;
    ata_read_sector(ATA_PRIMARY, ATA_MASTER, 0, page);
    printf("[KERNEL] Init program loaded to 0x%08X\n", page);
    printf("[KERNEL] Transferring control... goodbye\n");
    set_color(VGA_COLOR_WHITE);
    // 4 KiB for stack
    void *stack_bottom = mmap_find_first_free_block(&ctx);
    void *stack_top = stack_bottom + 0x1000;
    switch_to_user_mode(stack_top, page);
}

void kmain(struct multiboot_info *info) {
    clear_screen();
    set_color(VGA_COLOR_CYAN);
// Check if the bootloader gave us the upper and lower memory
    if (!(info->flags & 0x1)) goto halt;

    ctx.multi_mmap = (struct multiboot_mmap_entry *)info->memMapAddress;
    ctx.multi_mmap_size = info->memMapLength / sizeof(struct multiboot_mmap_entry);

    mmap_init(&ctx);
    acpi_init(&ctx);
    ps2_init();
    interrupts_init(g_idt, g_idtr);
    user_init();

halt:
    for (;;) {}
}
